import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        this._amount = 8;
        this._count = Math.pow(this._amount, 3);

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        //scene.background = new THREE.Color(0xffffff);
        this._scene = scene;

        this._setupCamera();
        //this._setupControls();
        //this._setupLight();
        this._setupModel();
        this._setupGUI();
        //this._setupHelper();
        
        window.onresize = this.resize.bind(this);

        this.resize();
    }

    _setupModel() {
        const loader = new THREE.BufferGeometryLoader()
        loader.load("models/json/suzanne_buffergeometry.json", (geometry) => {
            const instanceColors = [];
            for(let i=0; i<this._count; i++) {
                instanceColors.push(Math.random());
                instanceColors.push(Math.random());
                instanceColors.push(Math.random());
            }

            geometry.setAttribute("instanceColor", 
                new THREE.InstancedBufferAttribute(new Float32Array(instanceColors), 3));
            geometry.computeVertexNormals();
            geometry.scale(0.5, 0.5, 0.5);

            new THREE.TextureLoader().load("textures/matcaps/matcap-porcelain-white.jpg", (texture) => {
                texture.encoding = THREE.sRGBEncoding;
                const material = new THREE.MeshMatcapMaterial({ color: 0xaaaaff, matcap: texture }); 

                const colorParsChunk = [
                    'attribute vec3 instanceColor;',
                    'varying vec3 vInstanceColor;',
                    '#include <common>'
                ].join( '\n' );

                const instanceColorChunk = [
                    '#include <begin_vertex>',
                    '\tvInstanceColor = instanceColor;'
                ].join( '\n' );

                const fragmentParsChunk = [
                    'varying vec3 vInstanceColor;',
                    '#include <common>'
                ].join( '\n' );

                const colorChunk = [
                    'vec4 diffuseColor = vec4( diffuse * vInstanceColor, opacity );'
                ].join( '\n' );

                material.onBeforeCompile = function (shader) {
                    shader.vertexShader = shader.vertexShader
                        .replace( '#include <common>', colorParsChunk )
                        .replace( '#include <begin_vertex>', instanceColorChunk );
                    shader.fragmentShader = shader.fragmentShader
                        .replace( '#include <common>', fragmentParsChunk )
                        .replace( 'vec4 diffuseColor = vec4( diffuse, opacity );', colorChunk );
                    //console.log( shader.uniforms );
                    //console.log( shader.vertexShader );
                    //console.log( shader.fragmentShader );
                };

                const mesh = new THREE.InstancedMesh(geometry, material, this._count);
                mesh.instanceMatrix.setUsage( THREE.DynamicDrawUsage ); // will be updated every frame

                this._scene.add(mesh);
                this._mesh = mesh;
            });
        });
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        this._domWebGL.appendChild(stats.dom);
        this._stats = stats;
    }

    _setupCamera() {    
        const camera = new THREE.PerspectiveCamera(40, window.innerWidth / window.innerHeight, 1, 100);
        camera.position.set(0, 0, 20);

        this._camera = camera;
    }

    _setupControls() {
        //.
    }

    _setupLight() {
        //.
    }

    update() {
        if(this._mesh) {
            this._stats.update();

            const dummy = new THREE.Object3D();
            const time = Date.now() * 0.001;

            this._mesh.rotation.x = Math.sin(time / 4);
            this._mesh.rotation.y = Math.sin(time / 2);

            let i = 0;
            const offset = (this._amount - 1) / 2;
            for(let x=0; x<this._amount; x++) {
                for(let y=0; y<this._amount; y++) {
                    for(let z=0; z<this._amount; z++) {
                        dummy.position.set(offset-x, offset-y, offset-z);
                        dummy.rotation.y = (Math.sin(x/4+time) + Math.sin(y/4+time) + Math.sin(z/4+time));
                        dummy.rotation.z = dummy.rotation.y * 2;

                        dummy.updateMatrix();

                        this._mesh.setMatrixAt(i++, dummy.matrix);
                    }
                }
            }

            this._mesh.instanceMatrix.needsUpdate = true;
        }
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        
        this.update();
        this._renderer.render(this._scene, this._camera);
    }

    resize() {
        const aspect = window.innerWidth / window.innerHeight;

        this._camera.aspect = aspect;
        this._camera.updateProjectionMatrix();

        this._renderer.setPixelRatio( window.devicePixelRatio );
        this._renderer.setSize(window.innerWidth, window.innerHeight);
    }
}

window.onload = function () {
    (new App()).render();
}